"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
/*
   - index of hash-table is used as interned string

   - an entry can be viewed
     1. as a point
     2. as an orbit

   - open addressing
     for we do not need to delete
*/

function HASH_TABLE_ENTRY(index) {
  this.index = index;
  this.key = null;
  this.value = null;
  this.orbit_length = 0;
  this.orbiton = 0;
};

HASH_TABLE_ENTRY.prototype = {
  occured: function occured() {
    return this.key !== null;
  },

  used: function used() {
    return this.value !== null;
  },

  no_collision: function no_collision() {
    return this.index === this.orbiton;
  }
};

function HASH_TABLE(size, key_equal, hash) {
  this.size = size;
  this.key_equal = key_equal;
  this.hash = hash;
  this.array = new Array(size);
  this.counter = 0;
  for (var i = 0; i < size; i++) {
    this.array[i] = new HASH_TABLE_ENTRY(i);
  };
};

HASH_TABLE.prototype = {
  insert: function insert(key) {
    // key -> index
    //     -> null -- denotes the hash_table is filled
    var orbit_index = this.hash(key, 0);
    var counter = 0;
    while (true) {
      var index = this.hash(key, counter);
      var entry = this.index_to_entry(index);
      if (!entry.occured()) {
        entry.key = key;
        entry.orbiton = orbit_index;
        var orbit_entry = this.index_to_entry(orbit_index);
        orbit_entry.orbit_length = 1 + counter;
        this.counter = 1 + this.counter;
        return index;
      } else if (this.key_equal(key, entry.key)) {
        return index;
      } else if (counter === this.size) {
        return null;
      } else {
        counter = 1 + counter;
      }
    }
  },

  search: function search(key) {
    // key -> index
    //     -> null -- denotes key not occured
    var counter = 0;
    while (true) {
      var index = this.hash(key, counter);
      var entry = this.index_to_entry(index);
      if (!entry.occured()) {
        return null;
      } else if (this.key_equal(key, entry.key)) {
        return index;
      } else if (counter === this.size) {
        return null;
      } else {
        counter = 1 + counter;
      }
    }
  },

  key_to_index: function key_to_index(key) {
    var index = this.insert(key);
    if (index !== null) {
      return index;
    } else {
      console.log("hash_table is filled");
      throw "hash_table is filled";
    }
  },

  index_to_entry: function index_to_entry(index) {
    return this.array[index];
  },

  key_to_entry: function key_to_entry(key) {
    return index_to_entry(key_to_index(key));
  },

  report_orbit: function report_orbit(index, counter) {
    var entry = this.index_to_entry(index);
    while (true) {
      if (counter >= entry.orbit_length) {
        return;
      } else {
        var key = entry.key;
        var next_index = this.hash(key, counter);
        var next_entry = this.index_to_entry(next_index);
        if (index === next_entry.orbiton) {
          console.log("  - {", next_index, "} ", next_entry.key);
        }
        counter = 1 + counter;
      }
    }
  },

  report: function report() {
    console.log("\n");
    console.log("- hash_table-table report_used");
    var index = 0;
    while (true) {
      var entry = this.index_to_entry(index);
      if (index === this.size) {
        break;
      } else {
        if (entry.occured() && entry.no_collision()) {
          console.log("  - {", index, "} ", entry.key, " # ", entry.orbit_length);
          if (entry.used()) {
            console.log("      ", entry.value);
          }
          this.report_orbit(index, 1);
        }
        index = 1 + index;
      }
    }
    console.log("\n");
    console.log("- used : ", this.counter);
    console.log("- free : ", this.size - this.counter);
  }
};

exports.default = HASH_TABLE;