'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _jojoInJs = require('jojo-in-js');

var _jojoInJs2 = _interopRequireDefault(_jojoInJs);

var _HASH_TABLE = require('./HASH_TABLE.js');

var _HASH_TABLE2 = _interopRequireDefault(_HASH_TABLE);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var is = _jojoInJs2.default.is;
var BUFFER = _jojoInJs2.default.BUFFER;
var STACK = _jojoInJs2.default.STACK;
var argack = _jojoInJs2.default.argack;
var retack = _jojoInJs2.default.retack;
var v = _jojoInJs2.default.v;
var send = _jojoInJs2.default.send;
var recur = _jojoInJs2.default.recur;

var cell_area = new BUFFER(1024 * 1024, 0);

var argument_stack = new STACK(1024, 0);

function ARGUMENT_POINT(typetag, data) {
  this.typetag = typetag;
  this.data = data;
}

var scope_stack = new STACK(1024, 0);

function SCOPE_POINT() {}

var local_even_memory = new BUFFER(64 * 1024, 0);
var local_odd_memory = new BUFFER(64 * 1024, 0);

var return_stack = new STACK(1024, 0);

function RETURN_POINT() {}

var string_to_sum = function string_to_sum(string) {
  var sum = 0;
  var max_step = 10;
  for (var i = 0; i < string.length; i++) {
    sum = sum + string.codePointAt(i) * (2 << Math.min(i, max_step));
  }
  return sum;
};

var tag_hash_table = new _HASH_TABLE2.default(
// prime table size
997,
// key_equal
function (key1, key2) {
  return key1 === key2;
},
// hash
function (key, counter) {
  return (counter + string_to_sum(key)) % 997;
});

function array_equal(a, b) {
  if (a === b) return true;
  if (a == null || b == null) return false;
  if (a.length != b.length) return false;
  for (var i = 0; i < a.length; ++i) {
    if (a[i] !== b[i]) return false;
  }
  return true;
}

Array.prototype.last = function () {
  return this[this.length - 1];
};

function array_remove_dup(orig_array) {
  var array = orig_array.map(function (element) {
    return element;
  });
  array.sort();
  var result = [];
  for (var i = 0; i < array.length; i++) {
    if (result.last() === array[i]) {} else {
      result.push(array[i]);
    }
  }
  return result;
}

var tagset_hash_table = new _HASH_TABLE2.default(
// prime table size
997,
// key_equal
function (key1, key2) {
  return array_equal(array_remove_dup(key1), array_remove_dup(key2));
},
// hash
function (key, counter) {
  var sum = array_remove_dup(key).map(string_to_sum).reduce(function (sum, v) {
    return sum + v;
  });
  return (counter + sum) % 997;
});

var primitive_function_record = new BUFFER(1024, 0);

var create_primitive_function = function create_primitive_function(fun) {
  var return_address = primitive_function_record.cursor;
  primitive_function_record.set(return_address, fun);
  primitive_function_record.add_cursor(1);
  return return_address;
};

var address_after_explainer = 0;

var interpreter = function interpreter() {
  try {
    while (true) {
      var jojo = return_stack.pop();
      var jo = cell_area.get(jojo);
      var explainer = cell_area.get(jo);
      return_stack.push(jojo + 1);
      address_after_explainer = jo + 1;
      primitive_function_record.get(explainer).call();
    }
  } catch (string) {
    switch (string) {
      case "bye":
        break;
    }
  }
};

var in_host_tag_record = new Map();

var data = function data(value) {
  cell_area.set(cell_area.cursor, value);
  cell_area.add_cursor(1);
};

var mark = function mark(tag_string) {
  in_host_tag_record.set(tag_string, cell_area.cursor);
};

var link = 0;

var define_header = function define_header(tag_string, explainer) {
  data(link);
  link = cell_area.cursor - 1;
  mark(tag_string);
  data(explainer);
};

var define_primitive_function = function define_primitive_function(tag_string, fun) {
  var function_index = create_primitive_function(fun);
  define_header(tag_string, primitive_function_explainer);
  data(function_index);
};

var define_function = function define_function(tag_string, function_tag_string_array) {
  define_header(tag_string, function_explainer);
  function_tag_string_array.forEach(function (function_tag_string) {
    data(in_host_tag_record.get(function_tag_string));
  });
};

var define_variable = function define_variable(tag_string, value) {
  define_header(tag_string, variable_explainer);
  data(value);
};

var primitive_function_explainer = create_primitive_function(function () {
  primitive_function_record.get(cell_area.get(address_after_explainer)).call();
});

var function_explainer = create_primitive_function(function () {
  return_stack.push(address_after_explainer);
});

var variable_explainer = create_primitive_function(function () {
  argument_stack.push(cell_area.get(address_after_explainer));
});

define_primitive_function("end", function () {
  return_stack.pop();
});

define_primitive_function("bye", function () {
  console.log("bye bye ^-^/");
  throw "bye";
});

define_primitive_function("dup", function () {
  var a = argument_stack.pop();
  argument_stack.push(a);
  argument_stack.push(a);
});

define_primitive_function("mul", function () {
  var a = argument_stack.pop();
  var b = argument_stack.pop();
  argument_stack.push(a * b);
});

define_primitive_function("simple-wirte", function () {
  console.log(argument_stack.pop());
});

define_variable("little-test-number", 4);

define_function("square", ["dup", "mul", "end"]);

define_function("little-test", ["little-test-number", "square", "simple-wirte", "bye"]);

define_function("first-function", ["little-test", "end"]);

var begin_to_interpret_threaded_code = function begin_to_interpret_threaded_code() {
  var function_body_for_little_test = in_host_tag_record.get("first-function") + 1;
  return_stack.push(function_body_for_little_test);
  interpreter();
};

begin_to_interpret_threaded_code();

exports.default = {
  tag_hash_table: tag_hash_table,
  tagset_hash_table: tagset_hash_table
};