/*
   - index of hash-table is used as interned string

   - an entry can be viewed
     1. as a point
     2. as an orbit

   - open addressing
     for we do not need to delete
*/

function HASH_TABLE_ENTRY (index) {
  this.index = index;
  this.key = null;
  this.value = null;
  this.orbit_length = 0;
  this.orbiton = 0;
};

HASH_TABLE_ENTRY.prototype = {
  occured: function () {
    return this.key !== null;
  },

  used: function () {
    return this.value !== null;
  },

  no_collision: function () {
    return this.index === this.orbiton;
  }
};

function HASH_TABLE (size,
                     key_equal,
                     hash) {
  this.size = size;
  this.key_equal = key_equal;
  this.hash = hash;
  this.array = new Array(size);
  this.counter = 0;
  for (let i = 0; i < size ; i++) {
    this.array[i] = new HASH_TABLE_ENTRY(i);
  };
};

HASH_TABLE.prototype = {
  insert: function (key) {
    // key -> index
    //     -> null -- denotes the hash_table is filled
    let orbit_index = this.hash(key, 0);
    let counter = 0;
    while (true) {
      let index = this.hash(key, counter);
      let entry = this.index_to_entry(index);
      if (!entry.occured()) {
        entry.key = key;
        entry.orbiton = orbit_index;
        let orbit_entry = this.index_to_entry(orbit_index);
        orbit_entry.orbit_length = 1 + counter;
        this.counter = 1 + this.counter;
        return index;
      } else if (this.key_equal(key, entry.key)) {
        return index;
      } else if (counter === this.size) {
        return null;
      } else {
        counter = 1 + counter;
      }
    }
  },

  search: function (key) {
    // key -> index
    //     -> null -- denotes key not occured
    let counter = 0;
    while (true) {
      let index = this.hash(key, counter);
      let entry = this.index_to_entry(index);
      if (!entry.occured()) {
        return null;
      } else if (this.key_equal(key, entry.key)) {
        return index;
      } else if (counter === this.size) {
        return null;
      } else {
        counter = 1 + counter;
      }
    }
  },

  key_to_index: function (key) {
    let index = this.insert(key);
    if (index !== null) {
      return index;
    } else {
      console.log("hash_table is filled");
      throw "hash_table is filled";
    }
  },

  index_to_entry: function (index) {
    return this.array[index];
  },

  key_to_entry: function (key) {
    return index_to_entry(key_to_index(key));
  },

  report_orbit: function (index, counter) {
    let entry = this.index_to_entry(index);
    while (true) {
      if (counter >= entry.orbit_length) {
        return;
      } else {
        let key = entry.key;
        let next_index = this.hash(key, counter);
        let next_entry = this.index_to_entry(next_index);
        if (index === next_entry.orbiton) {
          console.log("  - {", next_index, "} ",
                      next_entry.key); }
        counter = 1 + counter;
      }
    }
  },

  report: function () {
    console.log("\n");
    console.log("- hash_table-table report_used");
    let index = 0;
    while (true) {
      let entry = this.index_to_entry(index);
      if (index === this.size) {
        break;
      } else {
        if (entry.occured() &&
            entry.no_collision()) {
          console.log("  - {", index, "} ",
                      entry.key, " # ",
                      entry.orbit_length);
          if (entry.used()) {
            console.log("      ", entry.value);
          }
          this.report_orbit(index, 1);
        }
        index = 1 + index;
      }
    }
    console.log("\n");
    console.log("- used : ", this.counter);
    console.log("- free : ", this.size - this.counter);
  }
};

export default HASH_TABLE;
